import zipfile
from django.core.exceptions import PermissionDenied
from django.http import HttpResponse
from django.http.response import Http404
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View
from .tools import get_lc_files_list
from .settings import TRANSTOOL_EXPORT_KEY, TRANSTOOL_PROJECT_BASE_DIR


class LocaleMessagesExportView(View):
    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        return super(LocaleMessagesExportView, self).dispatch(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        if not TRANSTOOL_EXPORT_KEY or request.POST.get('key') != TRANSTOOL_EXPORT_KEY:
            raise PermissionDenied
        if request.POST.get('po-only') == '1':
            exts = ['.po']
        elif request.POST.get('mo-only') == '1':
            exts = ['.mo']
        else:
            exts = ['.po', '.mo']
        response = HttpResponse(content_type='application/zip')
        response['Content-Disposition'] = 'attachment; filename="localemessages.zip"'
        z = zipfile.ZipFile(response, 'w', compression=zipfile.ZIP_DEFLATED)
        try:
            for abs_path, rel_path in get_lc_files_list(TRANSTOOL_PROJECT_BASE_DIR, exts=exts):
                with open(abs_path, 'rb') as tf:
                    z.writestr(rel_path, tf.read())
        finally:
            z.close()
        return response

    def get(self, request, *args, **kwargs):
        raise Http404



import base64
import hashlib
import os
import polib
import shutil
from io import BytesIO
from django.core import management
from django.core.management import BaseCommand, call_command
from django.utils.encoding import force_bytes
from ...tools import get_commonpath
from ...settings import TRANSTOOL_LOCALE_PATHS, TRANSTOOL_LOCALES, TRANSTOOL_DEFAULT_DOMAINS


try:  # todo add base makemessages command class into settings
    from django_jinja.management.commands.makemessages import Command as BaseMakemessagesCommand
except ImportError:
    from django.core.management.commands.makemessages import Command as BaseMakemessagesCommand


class CustomMakemessagesCommand(BaseMakemessagesCommand):
    source_dirs = None
    ignored_source_dirs = None
    update_all = None

    def add_arguments(self, parser):
        super(CustomMakemessagesCommand, self).add_arguments(parser)
        parser.add_argument('--add-source-dir', action='append', dest='source_dirs',
                            default=[], metavar='SOURCE_DIR',
                            help='Source directories to search phrases.')
        parser.add_argument('--update-all', action='store_true', dest='update_all', default=False,
                            help="Update all .po files even without any changes.")
        parser.add_argument('--ignore-source-dir', action='append', dest='ignored_source_dirs',
                            default=[], metavar='IGNORED_SOURCE_DIR',
                            help='Ignored source directories to skip searching phrases.')

    def handle(self, *args, **options):
        self.source_dirs = [os.path.abspath(p) for p in (options['source_dirs'] or [])]
        self.ignored_source_dirs = [os.path.abspath(p) for p in (options['ignored_source_dirs'] or [])]
        self.update_all = options['update_all']
        super(CustomMakemessagesCommand, self).handle(*args, **options)

    def find_files(self, root):
        files = []
        for source_dir in self.source_dirs:
            files.extend(super(CustomMakemessagesCommand, self).find_files(source_dir.rstrip('\\/')))

        if self.ignored_source_dirs:
            _files = []
            for f in files:
                ignore = False
                for ignored_source_dir in self.ignored_source_dirs:
                    if get_commonpath([ignored_source_dir]) == get_commonpath([ignored_source_dir, f.path]):
                        ignore = True
                        break
                if not ignore:
                    _files.append(f)
            files = _files
            del _files

        files.sort()
        return files

    @staticmethod
    def _get_hash_of_po_file(pofile_fn):
        h = hashlib.sha512()
        for entry in polib.pofile(pofile_fn):
            h.update(base64.b64encode(force_bytes(entry.msgid)))
            h.update(base64.b64encode(force_bytes(entry.msgstr)))
        return h

    def write_po_file(self, potfile, locale):
        pofile_fn = os.path.join(os.path.dirname(potfile), locale, 'LC_MESSAGES', '{}.po'.format(self.domain))
        old_po_hash = old_file = None
        if not self.update_all and os.path.isfile(pofile_fn):
            old_po_hash = self._get_hash_of_po_file(pofile_fn)
            old_file = BytesIO()
            with open(pofile_fn, 'rb') as f:
                shutil.copyfileobj(f, old_file)

        super(CustomMakemessagesCommand, self).write_po_file(potfile, locale)

        if old_po_hash:
            new_po_hash = self._get_hash_of_po_file(pofile_fn)
            if new_po_hash.hexdigest() == old_po_hash.hexdigest():
                old_file.seek(0)
                with open(pofile_fn, 'wb') as f:
                    shutil.copyfileobj(old_file, f)


def call_object_command(name, *args, **kwargs):
    def _get_command():
        return {name: name}

    old_function = management.get_commands
    try:
        management.get_commands = _get_command
        return call_command(name, *args, **kwargs)
    finally:
        management.get_commands = old_function


class Command(BaseCommand):
    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)
        parser.add_argument('--update-all', action='store_true', dest='update_all', default=False,
                            help="Update all .po files even without any changes.")

    def handle(self, *args, **options):
        for locale_path, locale_opts in TRANSTOOL_LOCALE_PATHS:
            locale_path = os.path.abspath(locale_path)
            os.chdir(locale_path)
            self.stdout.write('chdir {}'.format(locale_path))
            for domain in sorted(locale_opts.keys()):
                domain_opts = locale_opts[domain]
                command_kwargs = {
                    'domain': domain,
                    'extensions': domain_opts.get('EXT', TRANSTOOL_DEFAULT_DOMAINS[domain]['EXT']),
                    'source_dirs': domain_opts['DIRS'],
                    'locale': TRANSTOOL_LOCALES,
                    'update_all': options['update_all'],
                }
                if domain_opts.get('REST', False):
                    command_kwargs['ignored_source_dirs'] = self._get_excluded_source_dirs_for_rest(domain)
                call_object_command(CustomMakemessagesCommand(), **command_kwargs)

    @staticmethod
    def _get_excluded_source_dirs_for_rest(domain):
        dirs = []
        for locale_path, locale_opts in TRANSTOOL_LOCALE_PATHS:
            if domain not in locale_opts:
                continue
            domain_opts = locale_opts[domain]
            if domain_opts.get('REST', False):
                continue
            dirs.extend(domain_opts.get('DIRS', []))
        return dirs



from django.conf import settings


TRANSTOOL_DL_URL = getattr(settings, 'TRANSTOOL_DL_URL', None)  # http://example.com/localemessages/export/
TRANSTOOL_DL_KEY = getattr(settings, 'TRANSTOOL_DL_KEY', None)  # for import translates from remote server
TRANSTOOL_EXPORT_KEY = getattr(settings, 'TRANSTOOL_EXPORT_KEY', None)  # for export translates by http request
TRANSTOOL_PROJECT_BASE_DIR = getattr(settings, 'TRANSTOOL_PROJECT_BASE_DIR', settings.BASE_DIR)  # root dir of project
TRANSTOOL_LOCALES = getattr(settings, 'TRANSTOOL_LOCALES', tuple(map(lambda t: t[0], settings.LANGUAGES)))

TRANSTOOL_DEFAULT_DOMAINS = {
    'django': {
        'EXT': ('html', 'txt', 'py'),
    },
    'djangojs': {
        'EXT': ('js',),
    },
}

# update TRANSTOOL_DEFAULT_DOMAINS from project settings using addition method
for _domain, _opts in getattr(settings, 'TRANSTOOL_DEFAULT_DOMAINS', {}).items():
    if _domain in TRANSTOOL_DEFAULT_DOMAINS:
        # if domain is exists
        # just update options without remove default ones
        for _k, _v in _opts.items():
            TRANSTOOL_DEFAULT_DOMAINS[_domain][_k] = _v
    else:
        # add new domain
        TRANSTOOL_DEFAULT_DOMAINS[_domain] = _opts

# Example:
# TRANSTOOL_LOCALE_PATHS = (
#     (os.path.join(BASE_DIR, 'apps/myapp/'), {  # directory that contains locale subdirectory
#         'django': {
#             'DIRS': (
#                 os.path.join(BASE_DIR, 'apps/myapp/'),
#                 os.path.join(BASE_DIR, 'templates/myapp/'),
#             ),
#             # 'EXT': ('html', 'txt', 'py'),
#         },
#         'djangojs': {
#             'DIRS': (
#                 os.path.join(BASE_DIR, 'static/myapp/'),
#             ),
#             # 'EXT': ('js',),
#         },
#     }),
#     (os.path.join(BASE_DIR, './'), {
#         'django': {
#             'DIRS': (
#                 BASE_DIR,
#             ),
#             'REST': True,
#         },
#         'djangojs': {
#             'DIRS': (
#                 BASE_DIR,
#             ),
#             'REST': True,
#         },
#     }),
# )
TRANSTOOL_LOCALE_PATHS = getattr(settings, 'TRANSTOOL_LOCALE_PATHS', ())



import os
import sys
import tempfile
import zipfile
import requests
from io import BytesIO
from django.core.management import CommandError, BaseCommand
from ...tools import get_lc_files_list, get_diff_po
from ...settings import TRANSTOOL_DL_URL, TRANSTOOL_DL_KEY, TRANSTOOL_PROJECT_BASE_DIR


class Command(BaseCommand):
    help = ('Import locale messages from remote server. '
            '(You need configure TRANSTOOL_DL_URL and TRANSTOOL_DL_KEY settings)')

    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)
        parser.add_argument('--diff', action='store_true', default=False,
                            help='Only show difference.')
        parser.add_argument('--mo-only', action='store_true', default=False,
                            help='Import only mo-files.')
        parser.add_argument('--po-only', action='store_true', default=False,
                            help='Import only po-files.')

    @staticmethod
    def _get_diff_info(imp_zip_file, exts):
        result = {
            'changed': {},  # {rel path: number of changes}
            'new': [],  # rel path
            'del': [],  # rel path
            'ok': [],  # rel path
        }
        z = zipfile.ZipFile(imp_zip_file)
        try:
            for z_rel_path in z.namelist():
                abs_path = os.path.join(TRANSTOOL_PROJECT_BASE_DIR, z_rel_path)
                if not os.path.exists(abs_path):
                    result['new'].append(z_rel_path)
                else:
                    remote_file_content = z.read(z_rel_path)
                    ext = os.path.splitext(z_rel_path)[1]
                    if ext == '.po':
                        tmp_fd, remote_file_tmp_fn = tempfile.mkstemp(suffix='transtool')
                        try:
                            open(remote_file_tmp_fn, 'wb').write(remote_file_content)
                            d = get_diff_po(remote_file_tmp_fn, abs_path)
                            if d:
                                result['changed'][z_rel_path] = d
                            else:
                                result['ok'].append(z_rel_path)
                        finally:
                            os.close(tmp_fd)
                            os.remove(remote_file_tmp_fn)
                    elif ext == '.mo':
                        with open(abs_path) as f:
                            local_file_content = f.read()
                        if remote_file_content != local_file_content:
                            result['changed'][z_rel_path] = None
                        else:
                            result['ok'].append(z_rel_path)
        finally:
            z.close()
        for abs_path, rel_path in get_lc_files_list(TRANSTOOL_PROJECT_BASE_DIR, exts=exts):
            if rel_path not in result['changed'] and rel_path not in result['ok']:
                result['del'].append(rel_path)
        return result

    def print_diff_info(self, diff_info):
        self.stdout.write('New files:')
        if diff_info['new']:
            for fn in sorted(diff_info['new']):
                self.stdout.write('  + {}'.format(fn))
        else:
            self.stdout.write('  :empty:')

        self.stdout.write('Changed files:')
        if diff_info['changed']:
            for fn in sorted(diff_info['changed'].keys()):
                changes = diff_info['changed'][fn]
                if changes:
                    self.stdout.write('  * {} ({})'.format(fn, changes))
                else:
                    self.stdout.write('  * {}'.format(fn))
        else:
            self.stdout.write('  :empty:')

        self.stdout.write('Deleted files:')
        if diff_info['del']:
            for fn in sorted(diff_info['del']):
                self.stdout.write('  - {}'.format(fn))
        else:
            self.stdout.write('  :empty:')

    def copy_files(self, diff_info, imp_zip_file):
        z = zipfile.ZipFile(imp_zip_file)
        try:
            if diff_info['new']:
                self.stdout.write('Copy new files:')
                for fn in sorted(diff_info['new']):
                    content = z.read(fn)
                    try:
                        with open(os.path.join(TRANSTOOL_PROJECT_BASE_DIR, fn), 'wb') as f:
                            f.write(content)
                    except IOError as e:
                        self.stdout.write(e.message)
                        self.stdout.write('  SKIP: {}'.format(fn))
                    else:
                        self.stdout.write('  + {}'.format(fn))

            if diff_info['changed']:
                self.stdout.write('Copy changed files:')
                for fn in sorted(diff_info['changed'].keys()):
                    changes = diff_info['changed'][fn]
                    if changes:
                        fn_info = '{} ({})'.format(fn, changes)
                    else:
                        fn_info = fn
                    content = z.read(fn)
                    try:
                        with open(os.path.join(TRANSTOOL_PROJECT_BASE_DIR, fn), 'wb') as f:
                            f.write(content)
                    except IOError as e:
                        self.stdout.write(e.message)
                        self.stdout.write('  SKIP: {}'.format(fn_info))
                    else:
                        self.stdout.write('  * {}'.format(fn_info))

            if diff_info['del']:
                self.stdout.write('Maybe you should remove files:')
                for fn in sorted(diff_info['del']):
                    self.stdout.write('  - {}'.format(fn))
        finally:
            z.close()

    def handle(self, *args, **options):
        if not TRANSTOOL_DL_URL or not TRANSTOOL_DL_KEY:
            raise CommandError('Please, set TRANSTOOL_DL_URL and TRANSTOOL_DL_KEY settings.')
        if options['mo_only'] and options['po_only']:
            raise CommandError('Use only --mo-only or --po-only but not both.')
        self.stdout.write('Download file: Send POST request to {}'.format(TRANSTOOL_DL_URL))
        r = requests.post(TRANSTOOL_DL_URL, {
            'key': TRANSTOOL_DL_KEY,
            'po-only': str(int(options['po_only'])),
            'mo-only': str(int(options['mo_only'])),
        }, stream=True)
        if r.status_code != 200:
            self.stdout.write('Request status code is not 200: {}'.format(r.status_code))
            self.stdout.write('Fail.', ending='\n\n')
            sys.exit(1)
        file_content = BytesIO()
        for chunk in r.iter_content(chunk_size=(16 * 1024)):
            file_content.write(chunk)
        file_content.seek(0, os.SEEK_END)
        file_content_size = file_content.tell()
        self.stdout.write('Downloaded file {} {} bytes'.format(r.headers['Content-Type'], file_content_size))
        if options['po_only']:
            exts = ['.po']
        elif options['mo_only']:
            exts = ['.mo']
        else:
            exts = ['.po', '.mo']
        diff_info = self._get_diff_info(file_content, exts)
        if options['diff']:
            self.print_diff_info(diff_info)
        else:
            self.copy_files(diff_info, file_content)
        self.stdout.write('Done.', ending='\n\n')



import os
from distutils.core import setup
from setuptools import find_packages
import transtool


setup(
    name='django-transtool',
    version=transtool.__version__,
    description='Make translating your django project easier.',
    long_description=open(os.path.join(os.path.dirname(__file__), 'README.rst')).read(),
    license='MIT',
    author='Igor Melnyk @liminspace',
    author_email='liminspace@gmail.com',
    url='https://github.com/liminspace/django-transtool',
    packages=find_packages(exclude=('tests', 'tests.*')),
    include_package_data=True,
    zip_safe=False,  # because include static
    install_requires=[
        'django>=1.8,<1.12',
        'django-rosetta>=0.7.13',
        'requests>=2.8.1',
    ],
    keywords=[
        'django', 'django-transtool', 'translation', 'localization', 'internationalization', 'i18n', 'l10n',
    ],
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Framework :: Django',
        'License :: OSI Approved :: MIT License',
    ],
)
